@@ -3,6 +3,7 @@ This file is part of Swapspace.
 
 Copyright (C) 2005,2006, Software Industry Industry Promotion Agency (SIPA)
 Copyright (C) 2010, Jeroen T. Vermeulen
+Copyright (C) 2019, Jacob Adams
 Written by Jeroen T. Vermeulen.
 
 Swapspace is free software; you can redistribute it and/or modify
@@ -37,6 +38,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 #include <sys/statvfs.h>
 #include <sys/swap.h>
 #include <sys/types.h>
+#include <sys/ioctl.h>
+#include <sys/vfs.h>
+#include <linux/fs.h>
+#include <linux/magic.h>
 
 #include "log.h"
 #include "opts.h"
@@ -176,6 +181,26 @@ bool to_swapdir(void)
     return false;
   }
 #endif
+  struct stat swapdirstat;
+  mode_t mode = S_IFDIR | S_IRUSR | S_IWUSR | S_IXUSR;
+  if (stat(swappath, &swapdirstat) < 0)
+  {
+    log_perr(LOG_ERR, "Unable to stat swap path", errno);
+    return false;
+  }
+  if (swapdirstat.st_mode != mode)
+  {
+    logm(LOG_WARNING, "Swap file directory permissions wrong, fixing");
+    if (verbose)
+    {
+      logm(LOG_DEBUG, "Actual: %o\nExpected: %o", swapdirstat.st_mode, mode);
+    }
+    if (chmod(swappath, mode) < 0)
+    {
+      log_perr(LOG_ERR, "Unable to change swap file directory permissions", errno);
+      return false;
+    }
+  }
 
   return true;
 }
@@ -402,6 +427,52 @@ static memsize_t fill_swapfile(const char file[], int fd, memsize_t size)
   return bytes;
 }
 
+/// Ensure COW updates don't apply to swapfiles.
+/* Only needed to BTRFS and other COW filesystems.
+ * But can't hurt to apply everywhere.
+ */
+int set_no_cow(int fd)
+{
+  int attr;
+  int err;
+
+  err = ioctl(fd, FS_IOC_GETFLAGS, &attr);
+  if (err != 0) return err;
+  attr |= FS_NOCOW_FL;
+  err = ioctl(fd, FS_IOC_SETFLAGS, &attr);
+  if (err != 0) return err;
+  return 0;
+}
+
+/* Store filesystem type so we don't check everytime we allocate a swapfile
+ * __fsword_t is a glibc-internal type, but the docs claim unsigned int is fine
+ */
+unsigned int fstype = 0;
+
+/// Handle filesystem-specific operations needed for swapfiles
+/* Some filesystems, like BTRFS, require special operations to be performed on
+ * files before they can be used as swapfiles
+ * Clobbers localbuf if special operations are performed
+ */
+void specialfs(const char path[])
+{
+  int err;
+  /* All swapfiles exist on the same filesystem, so we should only need to
+   * call this once
+   */
+  if (fstype == 0)
+  {
+    struct statfs buf;
+    err = statfs(path, &buf);
+    if (err == 0) fstype = buf.f_type;
+    else log_perr_str(LOG_WARNING, "Could not detect filesystem", path, errno);
+  }
+  else if (fstype == BTRFS_SUPER_MAGIC)
+  {
+    err = runcommandformat("%s property set '%s' compression none", "btrfs", path);
+    if (err != 0) logm(LOG_WARNING, "Could not disable BTRFS compression! Return Code: %d", err);
+  }
+}
 
 /// Create file to be used as swap.  Clobbers localbuf.
 /**
@@ -430,6 +501,11 @@ static memsize_t make_swapfile(const char file[], memsize_t size)
     return 0;
   }
 
+  int err = set_no_cow(fd);
+  if (unlikely(err != 0)) log_perr_str(LOG_WARNING, "Could not disable COW", file, errno);
+
+  specialfs(file);
+
   if (unlikely(!fill_swapfile(file, fd, size)))
   {
     size = 0;
